/*
 *  Native support for the freeflow USB mouse.
 *
 *  Copyright (c) 2014      Pete Arnold      <peter.arnold@adder.com>
 *
 *  based on aiptek.c by
 *     Chris Atenasio      <chris@crud.net>
 *     Bryan W. Headley    <bwheadley@earthlink.net> 
 *  based on wacom.c by
 *     Vojtech Pavlik      <vojtech@suse.cz>
 *     Andreas Bach Aaen   <abach@stofanet.dk>
 *     Clifford Wolf       <clifford@clifford.at>
 *     Sam Mosel           <sam.mosel@computer.org>
 *     James E. Blair      <corvus@gnu.org>
 *     Daniel Egger        <egger@suse.de>
 *
 *  Many thanks to Oliver Kuechemann for his support.
 *
 *  ChangeLog:
 *
 * NOTE:
 *      This kernel driver is augmented by the "freeflow" XFree86 input
 *      driver for your X server.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/jiffies.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb/input.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>

/* Version Information
 */
#define DRIVER_VERSION "F1.2 (2.12.2014/13:50)"
#define DRIVER_AUTHOR  "Pete Arnold/Adder Technology Ltd"
#define DRIVER_DESC    "Free-Flow USB Mouse Driver (Linux 2.6.x)"

#define USB_VENDOR_ID_ADDER			0x21d1

/* Length of incoming packet from the tablet
 */
#define FREEFLOW_PACKET_LENGTH			8

struct freeflow_features {
	char usbPath[64 + 1];	/* device's physical usb path     */
	char inputPath[64 + 1];	/* input device path              */
};

struct freeflow {
	struct input_dev *inputdev;	   /* input device struct           */
	struct usb_device *usbdev;	   /* usb device struct             */
	struct urb *urb;		   /* urb for incoming reports      */
	dma_addr_t data_dma;		   /* our dma stuffage              */
	struct freeflow_features features; /* tablet's array of features    */
	unsigned int ifnum;		   /* interface number for IO       */
	int diagnostic;			   /* tablet diagnostic codes       */
	unsigned long eventCount;	   /* event count                   */
	unsigned char *data;		   /* incoming packet data          */
};

/***********************************************************************
 * Handle the URBs.
 */
static void freeflow_irq(struct urb *urb, struct pt_regs *regs)
{
	struct freeflow *ff = urb->context;
	unsigned char *data = ff->data;
	struct input_dev *inputdev = ff->inputdev;
	int retval, x, y, screen;

	printk(KERN_INFO "freeflow - irq (data[0]=%3d[1]=%3d[2]=%3d[3]=%3d[4]=%3d[5]=%3d).\n", 
				data[0], data[1], data[2], data[3], data[4], data[5]);
	//printk(KERN_INFO "urb transfer buffer length=%d.\n",
	//			urb->transfer_buffer_length);
	//printk(KERN_INFO "freelow - irq [0]=%d, [1]=%d.\n",
	//data[0], data[1]);

	printk(KERN_INFO "freeflow - irq urb->status=%d.\n", urb->status);
	switch (urb->status) {
	case 0:
		/* Success */
		printk(KERN_INFO "freeflow - irq - urb->status OK.\n");
		break;
	case -EOVERFLOW:
		printk(KERN_INFO "freeflow - irq - ignoring EOVERFLOW.\n");
		break;
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		/* This urb is terminated, clean up */
		dbg("%s - urb shutting down with status: %d",
		    __FUNCTION__, urb->status);
		return;

	default:
		dbg("%s - nonzero urb status received: %d",
		    __FUNCTION__, urb->status);
		goto exit;
	}

	printk(KERN_INFO "freeflow - irq - checking reports ...\n");

	ff->eventCount++;

	if (data[0] == 99) {
		input_regs(inputdev, regs);
		x = le16_to_cpu(get_unaligned((__le16 *) (data + 1)));
		y = le16_to_cpu(get_unaligned((__le16 *) (data + 3)));
		screen = data[5];
		printk(KERN_INFO "freeflow - irq - 99->input as %d@(%d, %d).\n",
					screen, x, y);
		input_report_abs(inputdev, ABS_X, x);
		input_report_abs(inputdev, ABS_Y, y);
		input_report_abs(inputdev, ABS_MISC, screen);
		input_sync(inputdev);
	} else {
		dbg("Unknown report %d", data[0]);
	}

exit:
	retval = usb_submit_urb(urb, GFP_ATOMIC);
	if (retval != 0) {
		err("%s - usb_submit_urb failed with result %d",
		    __FUNCTION__, retval);
	}
}

/***********************************************************************
 * The Freeflow VID/PIDs.
 */
static const struct usb_device_id freeflow_ids[] = {
	{USB_DEVICE(USB_VENDOR_ID_ADDER, 0x01)},
	{}
};

MODULE_DEVICE_TABLE(usb, freeflow_ids);

/***********************************************************************
 * Open an instance of the freeflow driver.
 */
static int freeflow_open(struct input_dev *inputdev)
{
	struct freeflow *ff = inputdev->private;

	ff->urb->dev = ff->usbdev;
	if (usb_submit_urb(ff->urb, GFP_KERNEL) != 0)
		return -EIO;

	printk(KERN_INFO "freeflow - open OK.\n");
	
	return 0;
}

/***********************************************************************
 * Close an instance of the freeflow driver.
 */
static void freeflow_close(struct input_dev *inputdev)
{
	struct freeflow *ff = inputdev->private;

	printk(KERN_INFO "freeflow - close.\n");

  	usb_kill_urb(ff->urb);
}

/***********************************************************************
 * This routine is called when freeflow has been identified. It basically
 * sets up the device and the driver's internal structures.
 */
static int
freeflow_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *usbdev = interface_to_usbdev(intf);
	struct usb_endpoint_descriptor *endpoint;
	struct freeflow *ff;
	struct input_dev *inputdev;
	struct input_handle *inputhandle;
	struct list_head *node, *next;

	printk(KERN_INFO "freeflow - probe.\n");

	ff = kzalloc(sizeof(struct freeflow), GFP_KERNEL);
	inputdev = input_allocate_device();
	if (!ff || !inputdev)
		goto fail1;

	ff->data = usb_buffer_alloc(usbdev, FREEFLOW_PACKET_LENGTH,
					SLAB_ATOMIC, &ff->data_dma);
	if (!ff->data)
		goto fail1;

	ff->urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!ff->urb)
		goto fail2;

	ff->inputdev = inputdev;
	ff->usbdev = usbdev;
	ff->ifnum = intf->altsetting[0].desc.bInterfaceNumber;

	/* Determine the usb devices' physical path.
	 * Asketh not why we always pretend we're using "../input0",
	 * but I suspect this will have to be refactored one
	 * day if a single USB device can be a keyboard & a mouse
	 * & a tablet, and the inputX number actually will tell
	 * us something...
	 */
	usb_make_path(usbdev, ff->features.usbPath,
			sizeof(ff->features.usbPath));
	strlcat(ff->features.usbPath, "/input0",
		sizeof(ff->features.usbPath));

	/* Set up client data, pointers to open and close routines
	 * for the input device.
	 */
	inputdev->name = "FreeflowUSBMouse";
	inputdev->phys = ff->features.usbPath;
	usb_to_input_id(usbdev, &inputdev->id);
	inputdev->cdev.dev = &intf->dev;
	inputdev->private = ff;
	inputdev->open = freeflow_open;
	inputdev->close = freeflow_close;

	/* Now program the capacities of the tablet, in terms of being
	 * an input device.
	 */
//	inputdev->evbit[0] |= BIT(EV_KEY)
//	    | BIT(EV_ABS)
//	    | BIT(EV_REL)
//	    | BIT(EV_MSC);
//
//	inputdev->absbit[0] |= BIT(ABS_MISC);
//
//	inputdev->relbit[0] |=
//	    (BIT(REL_X) | BIT(REL_Y) | BIT(REL_WHEEL) | BIT(REL_MISC));
//
//	inputdev->keybit[LONG(BTN_LEFT)] |=
//	    (BIT(BTN_LEFT) | BIT(BTN_RIGHT) | BIT(BTN_MIDDLE));
//
//	inputdev->keybit[LONG(BTN_DIGI)] |=
//	    (BIT(BTN_TOOL_PEN) |
//	     BIT(BTN_TOOL_RUBBER) |
//	     BIT(BTN_TOOL_PENCIL) |
//	     BIT(BTN_TOOL_AIRBRUSH) |
//	     BIT(BTN_TOOL_BRUSH) |
//	     BIT(BTN_TOOL_MOUSE) |
//	     BIT(BTN_TOOL_LENS) |
//	     BIT(BTN_TOUCH) | BIT(BTN_STYLUS) | BIT(BTN_STYLUS2));
//
//	inputdev->mscbit[0] = BIT(MSC_SERIAL);

	inputdev->evbit[0] |= BIT(EV_ABS);
	inputdev->absbit[0] |= BIT(ABS_X) | BIT(ABS_Y) | BIT(ABS_MISC);

	input_set_abs_params(inputdev, ABS_X, 0, 32767, 0, 0);
	input_set_abs_params(inputdev, ABS_Y, 0, 32767, 0, 0);
	input_set_abs_params(inputdev, ABS_MISC, 0, 1, 0, 0); /* Pta: added. */
	endpoint = &intf->altsetting[0].endpoint[0].desc;

	/* Go set up our URB, which is called when the device receives input.
	 */
	usb_fill_int_urb(ff->urb,
			 ff->usbdev,
			 usb_rcvintpipe(ff->usbdev,
					endpoint->bEndpointAddress),
			 ff->data, 8, freeflow_irq, ff,
			 endpoint->bInterval);

	ff->urb->transfer_dma = ff->data_dma;
	ff->urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	input_register_device(ff->inputdev);

	/* We now will look for the evdev device which is mapped to
	 * the freeflow mouse. The partial name is kept in the link list of
	 * input_handles associated with this input device.
	 * What identifies an evdev input_handler is that it begins
	 * with 'event', continues with a digit, and that in turn
	 * is mapped to input/eventN.
	 */
	list_for_each_safe(node, next, &inputdev->h_list) {
		inputhandle = to_handle(node);
		if (strncmp(inputhandle->name, "event", 5) == 0) {
			strcpy(ff->features.inputPath, inputhandle->name);
			break;
		}
	}

	/* Associate this driver's struct with the usb interface.
	 */
	usb_set_intfdata(intf, ff);

	/* Make sure the evdev module is loaded. Assuming evdev IS a module :-)
	 */
	if (request_module("evdev") != 0)
		info("ff: error loading 'evdev' module");

	return 0;

fail2:	usb_buffer_free(usbdev, FREEFLOW_PACKET_LENGTH, ff->data,
			ff->data_dma);
fail1:	input_free_device(inputdev);
	kfree(ff);
	return -ENOMEM;
}

/* Forward declaration */
static void freeflow_disconnect(struct usb_interface *intf);

static struct usb_driver freeflow_driver = {
	.name = "freeflow",
	.probe = freeflow_probe,
	.disconnect = freeflow_disconnect,
	.id_table = freeflow_ids,
};

/***********************************************************************
 * Deal with freeflow disconnecting from the system.
 */
static void freeflow_disconnect(struct usb_interface *intf)
{
	struct freeflow *ff = usb_get_intfdata(intf);

	printk(KERN_INFO "freeflow - disconnect.\n");
	
	/* Disassociate driver's struct with usb interface
	 */
	usb_set_intfdata(intf, NULL);
	if (ff != NULL) {
		/* Free & unhook everything from the system.
		 */
		usb_kill_urb(ff->urb);
		input_unregister_device(ff->inputdev);
		usb_free_urb(ff->urb);
		usb_buffer_free(interface_to_usbdev(intf),
				FREEFLOW_PACKET_LENGTH,
				ff->data, ff->data_dma);
		kfree(ff);
	}
}

static int __init freeflow_init(void)
{
	int result = usb_register(&freeflow_driver);
	
	printk(KERN_INFO "freeflow kernel driver "DRIVER_VERSION" - init.\n");

	if (result == 0) {
		info(DRIVER_VERSION ": " DRIVER_AUTHOR);
		info(DRIVER_DESC);
	}
	return result;
}

static void __exit freeflow_exit(void)
{
	printk(KERN_INFO "freeflow - exit.\n");
	
	usb_deregister(&freeflow_driver);
}

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

module_init(freeflow_init);
module_exit(freeflow_exit);
