/*
 * xf86Freeflow
 *
 * This driver assumes Linux USB/HID support, available for USB devices.
 * 
 * Version 1.0, 28 Ionawr 2015, Pete Arnold for Adder Technology
 * 
 * Copyright 2014-15 by Pete Arnold for Adder Technology <peter.arnold@adder.com>
 *
 * Lineage: This driver is based on the xf86Aiptek driver which, in turn, is
 *          based on both the xf86HyperPen and xf86Wacom tablet drivers.
 *
 *      xf86Aiptek -- (c) 2003 Bryan W. Headley
 *
 *      xf86HyperPen -- modified from xf86Summa (c) 1996 Steven Lang
 *          (c) 2000 Roland Jansen
 *          (c) 2000 Christian Herzog (button & 19200 baud support)
 *
 *      xf86Wacom -- (c) 1995-2001 Frederic Lepied
 *
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Bryan W. Headley not be used in 
 * advertising or publicity pertaining to distribution of the software 
 * without specific, written prior permission.  Bryan W. Headley makes no 
 * representations about the suitability of this software for any purpose.  
 * It is provided "as is" without express or implied warranty.
 *
 * PETE ARNOLD AND ADDER TECHNOLOGY DISCLAIM ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL EITHER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTIONS, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
/*
 * Section "InputDevice"
 *      Identifier  "Name"
 *      Driver      "freeflow"
 *      Option      "Device"        "pathname" {/dev/input/event0}
 *      Option      "DebugLevel"    "int"
 * EndSection
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "xf86Freeflow.h"
#include <errno.h>
#include <string.h>
#include <math.h>

static InputDriverPtr freeflowDriver;
static int debugLevel = INI_debugLevel;
static int screenCount = SCREEN_COUNT_MAX;
static char* screens[SCREEN_COUNT_MAX];
static Monitor screen[SCREEN_COUNT_MAX];
static int xMax = 0;
static int xMin = 0;
static int yMax = 0;
static int yMin = 0;

_X_EXPORT InputDriverRec FREEFLOW =
{
    1,                  /* driver version */
    "freeflow",         /* driver name */
    NULL,               /* identify */
    xf86FreeflowInit,   /* pre-init */
    xf86FreeflowUninit, /* un-init */
    NULL,               /* module */
    0                   /* ref count */
};

/*******************************************************************************
 * xf86FreeflowConvert
 * Convert valuators to X and Y. We deal with multiple X screens, adjusting
 * for xTop/xBottom/yTop/yBottom (or xSize/ySize).
 */
static Bool
xf86FreeflowConvert(LocalDevicePtr local,
                    int            first,
                    int            num,
                    int            v0,
                    int            v1,
                    int            v2,
                    int            v3,
                    int            v4,
                    int            v5,
                    int*           x,
                    int*           y)
{
    FreeflowDevicePtr device = (FreeflowDevicePtr)local->private;
    int screenToUse;
//#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    ScreenPtr pScreen = miPointerCurrentScreen();
//#else
//    ScreenPtr pScreen = miPointerGetScreen(local->dev);
//#endif

    DBG(D_INF, ErrorF("Freeflow Convert.\n"));
    DBG(D_DBG, ErrorF("Freeflow Convert enter with: v0=%d, v1=%d, pScreen=%d, current.scr=%d.\n",
        v0, v1, pScreen->myNum, device->current.scr));
    if (first != 0 || num == 1)
    {
        return FALSE;
    }
    /* Note that this is an adjustment to monitor co-ordinates where width and
     * height are the display resolution, v* and *Size are in the 32768 x 32768
     * range.
     */  
    if (device->current.scr != device->previous.scr)
    {
        DBG(D_DBG, ErrorF("Freeflow Convert moving screen cur.scr=%d, prv.scr=%d.\n",
            device->current.scr, device->previous.scr));
        screenToUse = device->previous.scr + 1;
    }
    else
    {
        screenToUse = device->current.scr + 1;
    }
    *x = ((v0 * xMax) / 32768) - screen[screenToUse].x;
    *y = ((v1 * yMax) / 32768) - screen[screenToUse].y;
    if (device->current.scr != device->previous.scr)
    {
        NewCurrentScreen(pScreen, *x, *y);
    }
    DBG(D_DBG, ErrorF("Freeflow Convert exit with: x=%d, y=%d, screen=%d.\n",
        *x, *y, device->current.scr));
    return TRUE;
}

/*******************************************************************************
 * xf86FreeflowPostPosition
 * Send X the latest position event.
 */
static void
xf86FreeflowPostPosition(LocalDevicePtr local)
{
    FreeflowDevicePtr device = (FreeflowDevicePtr)local->private;
    int s, x, y;
    ScreenPtr pScreen = miPointerCurrentScreen();

    DBG(D_INF, ErrorF("Freeflow PostPosition.\n"));

    if ((device->current.x == device->previous.x) &&
        (device->current.y == device->previous.y) &&
	(device->current.scr == device->previous.scr))
    {
        return;
    }

    if ((xMax == 0) || (yMax == 0))
	return;

    /* The multi-monitor calculation. */
    s = device->current.scr + 1;

    DBG(D_DBG, ErrorF("Freeflow - PP current screen = %d.\n", s));
    DBG(D_DBG, ErrorF("         - (%d, %d) @ %d x %d.\n",
        screen[s].x, screen[s].y, screen[s].w, screen[s].h));

    x = ((screen[s].x * 32768) + (screen[s].w * device->current.x)) / xMax;
    y = ((screen[s].y * 32768) + (screen[s].h * device->current.y)) / yMax;
    xf86PostMotionEvent(local->dev, ABSOLUTE_FLAG, 0, 2, x, y);
    device->previous.x   = device->current.x;
    device->previous.y   = device->current.y;
    device->previous.scr = device->current.scr;

    DBG(D_DBG, ErrorF("Freeflow posted (%d,%d).\n", x, y));
}

/*******************************************************************************
 * xf86FreeflowHIDReadInput
 * Read events from the device, save x,y,scr data, post when we get a sync.
 */
static void
xf86FreeflowHIDReadInput(LocalDevicePtr local)
{
    FreeflowDevicePtr   device = (FreeflowDevicePtr)local->private;
    ssize_t             len;
    struct input_event* event;
    char                eventbuf[sizeof(struct input_event) * MAX_EVENTS];

    SYSCALL(len = read(local->fd, eventbuf, sizeof(eventbuf)));
    if (len <= 0)
    {
        DBG(D_ERR, ErrorF("Error reading freeflow mouse: %s\n",
	    strerror(errno)));
        return;
    }
    for (event=(struct input_event *)(eventbuf);
         event<(struct input_event *)(eventbuf+len);
         ++event)
    {
        switch (event->type)
        {
            case EV_ABS:
            {
                switch (event->code)
                {
                    case ABS_X:
                    {
			DBG(D_DBG, ErrorF("Freeflow - ReadInput X = %d.\n",
				event->value));
                        device->current.x = event->value;
                    }
                    break;
                    case ABS_Y:
                    {
			DBG(D_DBG, ErrorF("Freeflow - ReadInput Y = %d.\n",
				event->value));
                        device->current.y = event->value;
                    }
                    break;
                    case ABS_MISC:
                    {
			DBG(D_DBG, ErrorF("Freeflow - ReadInput SCR = %d.\n",
				event->value));
                        /* The Linux FREEFLOW HID driver sends
                         * the screen number through ABS_MISC.
                         */
                        device->current.scr = event->value;
                    }
                    break;
                }
            }
            break;
        }
	if (event->type == EV_SYN)
            xf86FreeflowPostPosition(local);
    }
}

/*******************************************************************************
 * xf86FreeflowHIDOpen 
 */
static Bool
xf86FreeflowHIDOpen(LocalDevicePtr local)
{
    FreeflowDevicePtr device = (FreeflowDevicePtr)local->private;
    char              name[256] = "Unknown";
    int               version;

    local->fd = xf86OpenSerial(local->options);
    if (local->fd == -1)
    {
        DBG(D_ERR, ErrorF("Freeflow HIDOpen error opening %s : %s\n", 
		device->name, strerror(errno)));
        return !Success;
    }
    else
    {
        ioctl(local->fd, EVIOCGNAME(sizeof(name)), name);
        DBG(D_INF, ErrorF("%s HID Name: \"%s\"\n", XCONFIG_PROBED, name));
        ioctl(local->fd, EVIOCGVERSION, &version);
        DBG(D_INF, ErrorF("%s HID Driver Version: %d.%d.%d\n", XCONFIG_PROBED,
            version>>16, (version>>8) & 0xff, version & 0xff));
        DBG(D_INF, ErrorF("%s HID Driver is using %d as the fd\n",
            XCONFIG_PROBED, local->fd));
        return Success;
    }
}

/*******************************************************************************
 * xf86FreeflowOpenDevice
 * Opens and initializes the device driver.
 */
static int
xf86FreeflowOpenDevice(DeviceIntPtr pDriver)
{
    LocalDevicePtr    local  = (LocalDevicePtr)pDriver->public.devicePrivate;
    FreeflowDevicePtr device = (FreeflowDevicePtr)PRIVATE(pDriver);

    DBG(D_INF, ErrorF("Freeflow OpenDevice - fd=%d.\n", local->fd));
    if (local->fd < 0)
    {
        if (xf86FreeflowHIDOpen(local) != Success)
        {
            if (local->fd >= 0)
            {
               SYSCALL(close(local->fd));
            }
            local->fd = -1;
            DBG(D_WRN, ErrorF("Freeflow OpenDevice - failed to open.\n"));
            return !Success;
        }
    }
    /* Create axisStructs for every axis we support.
     * NOTE: min_resolution and max_resolution infers to
     * me a programmability to increase/decrease resolution.
     * We don't support that, so min & max = current_resolution.
     */
    InitValuatorAxisStruct(pDriver,                 /* X resolution */
               0,                                   /* axis_id */
               0,                                   /* min value */
               device->xBottom - device->xTop,      /* max value */
               LPI2CPM(375),                        /* resolution */
               LPI2CPM(375),                        /* min_resolution */
               LPI2CPM(375));                       /* max_resolution */

    InitValuatorAxisStruct(pDriver,                 /* Y Resolution */
               1,                                   /* axis_id */
               0,                                   /* min value */
               device->yBottom - device->yTop,      /* max value */
               LPI2CPM(375),                        /* resolution */
               LPI2CPM(375),                        /* min_resolution */
               LPI2CPM(375));                       /* max_resolution */
    return Success;
}

/*******************************************************************************
 * xf86FreeflowProc
 * Call dispatcher for this driver.
 */
static int
xf86FreeflowProc(DeviceIntPtr pFreeflow, int requestCode)
{
    int               numAxes = 2;
    LocalDevicePtr    local  = (LocalDevicePtr)pFreeflow->public.devicePrivate;
    FreeflowDevicePtr device = (FreeflowDevicePtr)PRIVATE(pFreeflow);

    switch (requestCode)
    {
        case DEVICE_INIT:
        {
            DBG(D_INF, ErrorF("Freeflow Proc request=INIT.\n"));
            if (InitValuatorClassDeviceStruct(pFreeflow,
                   numAxes,
                   xf86GetMotionEvents,
                   local->history_size,
                   Absolute) == FALSE)
            {
                DBG(D_ERR, ErrorF("Unable to allocate ValuatorClassDevice.\n"));
                return !Success;
            }
            xf86MotionHistoryAllocate(local);
            xf86FreeflowOpenDevice(pFreeflow);
        }
        break;
        case DEVICE_ON:
        {
            DBG(D_INF, ErrorF("Freeflow Proc request=ON.\n"));
            if (local->fd >= 0)
            {
                xf86AddEnabledDevice(local);
                pFreeflow->public.on = TRUE;
            }
            else
            {
                DBG(D_ERR, ErrorF("Unable to ON Freeflow, device not open.\n"));
                return !Success;
            }
        }
        break;
        case DEVICE_OFF:
        {
            DBG(D_INF, ErrorF("Freeflow Proc request=OFF.\n"));
            if (local->fd >= 0)
            {
                xf86FreeflowClose(local);
                xf86RemoveEnabledDevice(local);
            }
            pFreeflow->public.on = FALSE;
        }
        break;
        case DEVICE_CLOSE:
        {
            DBG(D_INF, ErrorF("Freeflow Proc request=CLOSE.\n")); 
            if (local->fd >= 0)
                xf86FreeflowClose(local);
        }
        break;
        default:
        {
            DBG(D_WRN, ErrorF("Freeflow Proc unsupported request=%d.\n",
		requestCode));
            return !Success;
        }
        break;
    }
    return Success;
}

/*******************************************************************************
 * xf86FreeflowClose
 */
static void
xf86FreeflowClose(LocalDevicePtr local)
{
    if (local->fd >= 0)
    {
        SYSCALL(close(local->fd));
    }
    local->fd = -1;
}

/*******************************************************************************
 * xf86FreeflowAllocate
 * Allocates the device structures for the FREEFLOW.
 */
static LocalDevicePtr
xf86FreeflowAllocate(void)
{
    LocalDevicePtr      local;
    FreeflowDevicePtr   device;
    int                 loop;

    DBG(D_INF, ErrorF("Freeflow Allocate.\n"));

    device = (FreeflowDevicePtr)xalloc(sizeof(FreeflowDeviceRec));
    if (!device)
    {
        DBG(D_ERR, ErrorF("Freeflow Allocate failed to allocate 'device'.\n"));
        return NULL;
    }
    local = xf86AllocateInput(freeflowDriver, 0);
    if (!local)
    {
        DBG(D_ERR, ErrorF("Freeflow Allocate failed at xf86AllocateInput.\n"));
        xfree(device);
        return NULL;
    }

    local->name            = "freeflow";
    local->type_name       = "freeflow";
    local->flags           = 0;
    local->device_control  = xf86FreeflowProc;
    local->read_input      = xf86FreeflowHIDReadInput;
    local->close_proc      = xf86FreeflowClose;
    local->conversion_proc = xf86FreeflowConvert;
    local->fd              = VALUE_NA;
    local->atom            = 0;
    local->dev             = NULL;
    local->private         = device;
    local->private_flags   = 0;
    local->history_size    = 0;

    device->name           = "";
    device->xTop           = 0;
    device->yTop           = 0;
    device->xBottom        = 32767;
    device->yBottom        = 32767;
    device->current.x      = 0;
    device->current.y      = 0;
    device->current.scr    = 0;
    device->previous.x     = 0;
    device->previous.y     = 0;
    device->previous.scr   = 0;

    screenCount = SCREEN_COUNT_MAX;
    for (loop = 0; loop < screenCount; loop++)
    {
        screens[loop] = "";
	screen[loop].x = screen[loop].y = 0;
        screen[loop].w = 1920;
        screen[loop].h = 1080;
    }

    return local;
}

/*******************************************************************************
 * xf86FreeflowUninit
 * Called when the driver is unloaded.
 */
static void
xf86FreeflowUninit(InputDriverPtr drv,
                   LocalDevicePtr local,
                   int            flags)
{
    FreeflowDevicePtr device = (FreeflowDevicePtr)local->private;
    
    DBG(D_INF, ErrorF("Freeflow Uninit.\n"));

    xf86FreeflowProc(local->dev, DEVICE_OFF);
    DBG(D_INF, ErrorF("Freeflow Uninit -- after Proc ==> OFF.\n"));
    xfree(device);
    DBG(D_INF, ErrorF("Freeflow Uninit -- after freeing device.\n"));
    xf86DeleteInput(local, 0);
    DBG(D_INF, ErrorF("Freeflow Uninit -- after xf86DeleteInput - exiting.\n"));
}

/*******************************************************************************
 * xf86FreeflowInit
 * Called when the module subsection is found in XF86Config.
 */
static InputInfoPtr
xf86FreeflowInit(InputDriverPtr drv,
                 IDevPtr        dev,
                 int            flags)
{
    LocalDevicePtr    local = NULL;
    FreeflowDevicePtr device = NULL;
    int               loop;
    char              nextScreen[16];
    char*             nextParameter = "";
    int               x, y, w, h;

    DBG(D_NOT, ErrorF("Freeflow Init - built on "__DATE__", "__TIME__".\n"));

    freeflowDriver = drv;
    local = xf86FreeflowAllocate();
    if(!local)
    {
        DBG(D_ERR, ErrorF("Freeflow Init - failed to allocate local data.\n"));
        return NULL;
    }
    local->conf_idev = dev;
    xf86CollectInputOptions(local, NULL, NULL);
    device = (FreeflowDevicePtr)local->private;
    local->name = dev->identifier;
    /* Process the options from xorg.conf. */
    /* Device (mandatory). */
    device->name = xf86FindOptionValue(local->options, "Device");
    if (!device->name)
    {
        DBG(D_WRN, ErrorF("Freeflow Init - No \"Device\" in xorg.conf.\n"));
        goto SetupProc_fail;
    }
    else
    {
        DBG(D_NOT, ErrorF("Freeflow Init - Device is %s.\n", device->name));
    }
    /* DebugLevel (optional).*/
    debugLevel = xf86SetIntOption(local->options, "DebugLevel", debugLevel);
    if (debugLevel > 0)
    {
        DBG(D_NOT, ErrorF("Freeflow Init - debug level set to %d.\n",
            debugLevel));
    }
    /* Screen definitions. */
    screenCount = SCREEN_COUNT_MAX;
    for (loop = 1; loop <= screenCount; loop++)
    {
        sprintf(nextScreen, "Screen%d", loop);

        DBG(D_DBG, ErrorF("Freeflow Init - looking for %s definition.\n",
            nextScreen));

        nextParameter = xf86FindOptionValue(local->options, nextScreen);

        if (!nextParameter)
        {
            screenCount = loop - 1;
            break;
        }
        else
        {
            screens[loop] = nextParameter;
        }
    }
    /* Process the screen data if there is any. */
    if (screenCount == 0)
    {
        screen[0].x = screen[0].y = 0;
        screen[0].w = xMax = 1920;
        screen[0].h = yMax = 1080;
    }
    else
    {   
        xMax = yMax = 0; 
        for (loop = 1; loop <= screenCount; loop++)
        {
            x = y = w = h = -1;
            nextParameter = screens[loop];
            x = atoi(nextParameter);
            if (x < 0) goto broken; 
            nextParameter = strchr(nextParameter, SEPARATOR) + 1;
            y = atoi(nextParameter);
            if (y < 0) goto broken;
            nextParameter = strchr(nextParameter, SEPARATOR) + 1;
            w = atoi(nextParameter);
            if (w < 0) goto broken;
            w = w - x;
            nextParameter = strchr(nextParameter, SEPARATOR) + 1;
            h = atoi(nextParameter);
            if (h < 0) goto broken;
            h = h - y;
            screen[loop].x = x;
            screen[loop].y = y;
            screen[loop].w = w;
            screen[loop].h = h;
            if ((x + w) > xMax) xMax = x + w;
            if ((y + h) > yMax) yMax = y + h;
            continue;
        broken:
            screenCount = loop - 1;
            break;
        }
        for (loop = 1; loop <= screenCount; loop++)
        {
            DBG(D_DBG, ErrorF("Freeflow Init - scr %d = (%d, %d) @ %d x %d.\n",
                loop, screen[loop].x, screen[loop].y, screen[loop].w,
                screen[loop].h));
        }
        DBG(D_DBG, ErrorF("Freeflow Init - xMax = %d, yMax = %d.\n",
            xMax, yMax));
    }
    /* Standard options. */
    xf86ProcessCommonOptions(local, local->options);
    local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;

    DBG(D_DBG, ErrorF("Freeflow Init - finished.\n"));

    return (local);

SetupProc_fail:
    DBG(D_DBG, ErrorF("Freeflow Init - failed.\n"));
    if (device)
        xfree(device);
    if (local)
        xfree(local);
    return NULL;
}

/*******************************************************************************
 * Dynamic loading functions
 */
#ifdef XFree86LOADER
/*******************************************************************************
 * xf86FreeflowUnplug
 * Called when the module subsection is found in XF86Config.
 */
static void
xf86FreeflowUnplug(pointer p)
{
    DBG(D_WRN, ErrorF("Freeflow Unplug.\n"));
}

/*******************************************************************************
 * xf86FreeflowPlug
 * Called when the module subsection is found in XF86Config.
 */
static pointer
xf86FreeflowPlug(pointer module,
                 pointer options,
                 int*    errmaj,
                 int*    errmin)
{
    DBG(D_WRN, ErrorF("Freeflow Plug.\n"));
    xf86AddInputDriver(&FREEFLOW, module, 0);
    return module;
}

static XF86ModuleVersionInfo xf86FreeflowVersionRec =
{
    "freeflow",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    1, 0, 3,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}        /* Signature, to be patched into the file later. */
};

_X_EXPORT XF86ModuleData freeflowModuleData =
{
    &xf86FreeflowVersionRec,
    xf86FreeflowPlug,
    xf86FreeflowUnplug
};
#endif /* XFree86LOADER */

/* end of xf86Freeflow.c */
